import itertools
import json
import sys
from functools import partial
from typing import Annotated, cast

import senkawa
import typer
from pydantic import BaseModel, Field, ValidationError
from returns.functions import raise_exception
from returns.pipeline import flow, is_successful
from returns.pointfree import bind
from returns.result import ResultE, safe

from azuma import schemas

app = typer.Typer()


class ScanResult(BaseModel):
    path: str = Field(...)
    matched: bool = Field(...)


@safe(exceptions=(ValidationError,))
def load_rule(path: str) -> schemas.Rule:
    return schemas.Rule.model_validate_file(path)


@app.command()
def scan(
    path: Annotated[
        str, typer.Argument(help="Path (or glob pattern) to rule YAML file(s)")
    ],
    target: Annotated[
        str, typer.Argument(help="Path (or glob pattern) to event JSON file(s)")
    ],
):
    @safe
    def load_rules(path: str) -> list[schemas.Rule]:
        results = [load_rule(path_) for path_ in senkawa.glob(path)]
        return [result.alt(raise_exception).unwrap() for result in results]

    @safe
    def scan(rules: list[schemas.Rule], *, target: str) -> dict[str, list[dict]]:
        memo: dict[str, list[dict]] = {}
        for rule in rules:
            results: list[ScanResult] = []

            for path in senkawa.glob(target):
                with open(path) as f:
                    data = json.loads(f.read())
                    results.append(ScanResult(path=path, matched=rule.match(data)))

            key = rule.id or rule.title
            memo[key] = [r.model_dump() for r in results]

        return memo

    @safe
    def output(results: dict[str, list[dict]]) -> None:
        print(json.dumps(results))  # noqa: T201

    result: ResultE[None] = flow(
        path, load_rules, bind(partial(scan, target=target)), bind(output)
    )
    result.alt(raise_exception)


@app.command()
def validate(
    path: Annotated[
        list[str],
        typer.Argument(help="Path(s) (or glob pattern(s)) to rule YAML file(s)"),
    ],
):
    memo: dict[str, ValidationError] = {}
    for path_ in itertools.chain.from_iterable([senkawa.glob(p) for p in path]):
        result: ResultE[schemas.Rule] = flow(path_, load_rule)
        if not is_successful(result):
            memo[path_] = cast(ValidationError, result.failure())

    if not memo:
        return

    for path_, error in memo.items():
        print(f"{path_} has {error}\n", file=sys.stderr)  # noqa: T201

    sys.exit(1)


if __name__ == "__main__":
    typer.run(app())



class AzumaError(Exception):
    pass


class UnsupportedFeatureError(AzumaError):
    """Raised when a signature using an unsupported sigma feature is loaded."""

    pass


class RuleLoadError(AzumaError):
    pass



import base64
from typing import Any

import regex as re

from azuma import types
from azuma.exceptions import UnsupportedFeatureError

# TODO We need to support the rest of them
SUPPORTED_MODIFIERS = {
    "contains",
    "all",
    "base64",
    # 'base64offset'
    "endswith",
    "startswith",
    # 'utf16le',
    # 'utf16be',
    # 'wide',
    # 'utf16',
    "re",
    # 'windash'
}


def decode_base64(x: str) -> str:
    x = x.replace("\n", "")
    return base64.b64encode(x.encode()).decode()


MODIFIER_FUNCTIONS = {
    "contains": lambda x: f".*{x}.*",
    "base64": lambda x: decode_base64(x),
    "endswith": lambda x: f".*{x}$",
    "startswith": lambda x: f"^{x}.*",
}


def process_field_name(field_string: str) -> tuple[str, list[str]]:
    name_and_modifiers = field_string.split("|")
    name = name_and_modifiers.pop(0)
    modifiers = [m for m in name_and_modifiers if m]

    unsupported = set(modifiers) - SUPPORTED_MODIFIERS
    if unsupported:
        raise UnsupportedFeatureError(
            f"Unsupported field modifiers used: {unsupported}"
        )

    return name, modifiers


_NSC = NON_SPECIAL_CHARACTERS = r"[^\\*?]*"
ESCAPED_SPECIAL_CHARACTER = r"(?:\\[*?])"
ESCAPED_OTHER_CHARACTER = r"(?:\\[^*?])"
ESCAPED_WILDCARD_PATTERN = re.compile(
    rf"(?:{_NSC}{ESCAPED_SPECIAL_CHARACTER}*{ESCAPED_OTHER_CHARACTER})*"
)

UPTO_WILDCARD = re.compile(r"^([^\\?*]+|(?:\\[^?*\\])+)+")


def sigma_string_to_regex(original_value: str) -> str:
    value = original_value
    full_content: list[str] = []

    while value:
        # Grab any content up to the first wildcard
        match = UPTO_WILDCARD.match(value)

        if match:
            # The non regex content in the sigma string, may have characters special to regex
            matched = match.group(0)
            full_content.append(re.escape(matched))
            value = value[len(matched) :]
            continue

        if value.startswith("*"):
            full_content.append(".*")
            value = value[1:]
            continue

        if value.startswith("\\*"):
            full_content.append(re.escape("*"))
            value = value[2:]
            continue

        if value.startswith("?"):
            full_content.append(".")
            value = value[1:]
            continue

        if value.startswith("\\?"):
            full_content.append(re.escape("?"))
            value = value[2:]
            continue

        if value.startswith(r"\\*"):
            full_content.append(re.escape("\\") + ".*")
            value = value[3:]
            continue

        if value.startswith(r"\\?"):
            full_content.append(re.escape("\\") + ".")
            value = value[3:]
            continue

        if value.startswith("\\"):
            full_content.append(re.escape("\\"))
            value = value[1:]
            continue

        raise ValueError(f"Could not parse string matching pattern: {original_value}")

    return "".join(full_content)  # Sigma strings are case insensitive


def get_modified_value(value: str, modifiers: list[str] | None) -> str:
    if not modifiers:
        # If there are no modifiers, we assume exact match
        return f"^{value}$"

    for mod in modifiers:
        func = MODIFIER_FUNCTIONS.get(mod)
        value = func(value) if func else value

    return value


MODIFIER_REGEX_FLAGS = re.IGNORECASE | re.V1 | re.DOTALL


def apply_modifiers(value: str, modifiers: list[str]) -> types.Query:
    """
    Apply as many modifiers as we can during signature construction
    to speed up the matching stage as much as possible.
    """

    # If there are wildcards, or we are using the regex modifier, compile the query
    # string to a regex pattern object
    has_re = "re" in modifiers
    has_multiple_modifiers = len(modifiers) > 1

    if has_re and has_multiple_modifiers:
        raise ValueError("re modifier cannot use along with other modifiers")

    if has_re:
        return re.compile(value, flags=MODIFIER_REGEX_FLAGS)

    if not ESCAPED_WILDCARD_PATTERN.fullmatch(value):
        # Transform the unescaped wildcards to their regex equivalent
        reg_value = sigma_string_to_regex(value)
        value = get_modified_value(reg_value, modifiers)
        return re.compile(value, flags=MODIFIER_REGEX_FLAGS)

    value = get_modified_value(value, modifiers)
    # If we are just doing a full string compare of a raw string, the comparison
    # is case-insensitive in sigma, so all direct string comparisons will be lowercase.
    value = str(value).replace("\\*", "*").replace("\\?", "?")
    return value.lower()


def normalize_field_map(field: dict[str, Any]) -> types.DetectionMap:
    def map_raw_key_value(raw_key: str, value: Any) -> types.DetectionItem:
        key, modifiers = process_field_name(raw_key)

        if value is None:
            return (key, ([None], modifiers))

        if isinstance(value, list):
            return (
                key,
                (
                    [
                        apply_modifiers(str(v), modifiers) if v is not None else None
                        for v in value
                    ],
                    modifiers,
                ),
            )

        return (key, ([apply_modifiers(str(value), modifiers)], modifiers))

    return [map_raw_key_value(raw_key, value) for raw_key, value in field.items()]



import re
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, Optional, Union

if TYPE_CHECKING:
    from . import schemas


Query = Optional[Union[str, re.Pattern, Any]]
DetectionItem = tuple[str, tuple[list[Query], list[str]]]
DetectionMap = list[DetectionItem]
Condition = Callable[["schemas.Rule", dict[Any, Any]], Any]



from pathlib import Path
from typing import Any

import yaml
from pydantic import BaseModel


class YamlBaseModel(BaseModel):
    """BaseModel with YAML support"""

    @classmethod
    def model_validate_file(  # type: ignore
        cls,
        path: str | Path,
        *,
        encoding: str = "utf8",
        strict: bool | None = None,
        context: dict[str, Any] | None = None,
    ):
        """Parse a YAML file.

        Args:
            path (str | Path): Path to a file.
            encoding (str, optional): Encoding. Defaults to "utf8".
            strict (bool | None, optional): Strict or not. Defaults to None.
            context (dict[str, Any] | None, optional): Context. Defaults to None.

        Returns:
            YamlBaseModel: Parsed instance.
        """
        with open(path, encoding=encoding) as f:
            text = f.read()

        obj = yaml.safe_load(text)
        return cls.model_validate(obj, strict=strict, context=context)

    @classmethod
    def model_validate_yaml(  # type: ignore
        cls,
        b: str | bytes,
        *,
        strict: bool | None = None,
        context: dict[str, Any] | None = None,
    ):
        """Parse a YAML text.

        Args:
            b (str | bytes): String or bytes.
            strict (bool | None, optional): Strict or not. Defaults to None.
            context (dict[str, Any] | None, optional): Context. Defaults to None.

        Returns:
            YamlBaseModel: Parsed instance.
        """
        obj = yaml.safe_load(b)
        return cls.model_validate(obj, strict=strict, context=context)



from pathlib import Path
from typing import Any

import senkawa
from pydantic import Field, RootModel

from .rule import Rule


class RuleSet(RootModel):
    root: list[Rule] = Field(default_factory=list)

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item):
        return self.root[item]

    def match_all(self, event: dict[Any, Any]) -> list[Rule]:
        """Check whether an event is matched with the rules

        Args:
            event (dict[Any, Any]): Event

        Returns:
            list[Rule]: A list of matched rules
        """
        return [rule for rule in self if rule.match(event)]

    def unique(self) -> "RuleSet":
        """Returns unique rule set.

        Returns:
            RuleSet: Rule set
        """
        seen: set[str] = set()

        filtered: list[Rule] = []
        for rule in self.root:
            if rule.id is None:
                filtered.append(rule)
                continue

            if rule.id not in seen:
                seen.add(rule.id)
                filtered.append(rule)

        return RuleSet(root=filtered)

    @classmethod
    def from_dir(cls, dir: str | Path, *, pattern="*.{yml,yaml}") -> "RuleSet":
        """Load rules from a directory

        Args:
            dir (str | Path): Directory
            pattern (str, optional): YAML file pattern. Defaults to "*.{yml,yaml}".

        Returns:
            RuleSet: Rule set
        """
        dir = Path(dir) if isinstance(dir, str) else dir
        expanded = senkawa.glob(str(dir.joinpath(pattern)))
        return cls(root=[Rule.model_validate_file(p) for p in expanded])



import fnmatch
from typing import TYPE_CHECKING, Any

import regex as re

from . import types

if TYPE_CHECKING:
    from . import schemas


def match_search_id(rule: "schemas.Rule", event: dict[Any, Any], search_id: str):
    search_fields = rule.detection.get_search_fields(search_id)
    if search_fields is None:
        raise ValueError(f"Search fields for {search_id} are missing")

    return find_matches(event, search_fields)


def check_pair(event: dict[Any, Any], key: str, value: types.Query) -> bool:
    """
    Checks to see if a given key and value from the rule are also in the event.
    Takes into consideration any value modifiers.

    :param event: dict, a single event from the event log
    :param key: str, given dict key
    :param value: str, given key value
    :return: bool, whether or not the match exists in the event
    """
    # Before we can apply modifiers and search properly, we need to check if there
    # is even a value to modify, so do the null checks first
    if value is None:
        return event.get(key) is None

    if key not in event:
        return False

    if isinstance(value, re.Pattern):
        return bool(value.match(str(event[key])))

    # Because by default sigma string matching is case insensitive, lower the event
    # string before comparing it. The value string is already lowercase.
    # TODO potential optimization by caching lowercased event fields
    return str(event[key]).lower() == value


def find_matches(
    event: dict[Any, Any], search: "schemas.DetectionField", match_all: bool = False
) -> bool:
    """
    Matches the items in the rule to the event. Iterates through the sections and if there's a list it iterates
    through that. Uses checkPair to see if the items in the list/dictionary match items in the event log.

    :param event: dict, event read from the Sysmon log
    :param search: An object describing what sort of search to run
    :param match_all: A bool indicating if we want all fields in list to hit
    :return: bool, whether or not we found a match
    """
    if search.list_search:
        check = all if match_all else any
        return check(
            any(check_pair(event, event_key, field) for event_key in event)
            for field in search.list_search
        )

    return any(find_matches_by_map(event, field) for field in search.map_search)


def find_matches_by_map(event: dict[Any, Any], search: types.DetectionMap) -> bool:
    """
    :param event:
    :param search: a dict of fields to search. All must be satisfied.
    :return:
    """
    for field_name, (value, modifiers) in search:
        if not find_matches_by_map_entry(event, field_name, value, modifiers):
            return False

    return True


def find_matches_by_map_entry(
    event: dict,
    field_name: str,
    field_values: list[types.Query],
    modifiers: list[str],
) -> bool:
    """
    :param event: the event to search in
    :param field_name: A field in the event we want to search
    :param field_values: valid values or patterns for the field in question
    :return:
    """
    # Normally any of the values in field_values is acceptable, but the all modifier inverts that
    if "all" in modifiers:
        for permitted_value in field_values:
            if not check_pair(event, field_name, permitted_value):
                return False

        return True

    for permitted_value in field_values:
        if check_pair(event, field_name, permitted_value):
            return True

    return False


def analyze_x_of(
    rule: "schemas.Rule",
    event: dict[Any, Any],
    count: int | None = None,
    selector: str | None = None,
) -> bool:
    """
    Analyzes the truth value of an 'x of' condition specified within the condition string of the rule.

    :param rule: Rule currently being applied
    :param event: event currently being scanned
    :param count: left side of the x of statement, either 1 or None (for all)
    :param selector: right side of the x of statement, a pattern or None (for all)
    :return: bool, truth value of 'x of' condition
    """
    # First we need to choose our set of fields based on our selector.
    matches: dict[str, "schemas.DetectionField"] = {}
    all_searches = rule.detection.all_searches

    if selector is None:  # None indicates all.
        matches = all_searches
    else:
        for search_id, search_fields in all_searches.items():
            if fnmatch.fnmatch(search_id, selector):
                matches[search_id] = search_fields

    match_all = False
    if count is None:
        match_all = True
        count = len(matches)

    permitted_misses = len(matches) - count

    # Now that we have our searches to check, run them
    search_hits = 0
    search_misses = 0
    for search_fields in matches.values():
        if find_matches(event, search_fields, match_all):
            search_hits += 1
        else:
            search_misses += 1

        # Short circuit if we found the matches, or if we can't find the number anymore
        if search_hits >= count:
            return True

        if search_misses > permitted_misses:
            return False

    return False
